@@ -1,6 +1,7 @@
 from pyscipopt import Model, Eventhdlr, SCIP_EVENTTYPE
 from sklearn import manifold
 import pandas as pd
+import numpy as np
 import plotly.graph_objs as go
 import networkx as nx
 import os
@@ -41,7 +42,10 @@ class LPstatEventhdlr(Eventhdlr):
         age = self.model.getNNodes()
         condition = math.log10(self.model.getCondition())
         iters = self.model.lpiGetIterations()
-
+        pb = self.model.getPrimalbound()
+        if pb >= self.model.infinity():
+            pb = None
+        
         nodedict = {
             "number": node.getNumber(),
             "LPsol": LPsol,
@@ -55,11 +59,13 @@ class LPstatEventhdlr(Eventhdlr):
             # "variables": self.model.getNVars(),
             # "constraints": self.model.getNConss(),
             "rows": self.model.getNLPRows(),
+            "primalbound": pb,
+            "dualbound": self.model.getDualbound(),
         }
         # skip 0-iterations LPs (duplicates?)
         if firstlp:
             self.nodelist.append(nodedict)
-        elif iters >= 0:
+        elif iters > 0:
             prevevent = self.nodelist[-1]
             if nodedict["number"] == prevevent["number"] and not prevevent["first"]:
                 # overwrite data from previous LP event
@@ -279,14 +285,26 @@ class TreeD:
             active=0,
             yanchor="top",
             xanchor="left",
-            currentvalue={"prefix": "Age:", "visible": True, "xanchor": "right",},
+            currentvalue={
+                "prefix": "Age:",
+                "visible": True,
+                "xanchor": "right",
+            },
             len=0.9,
             x=0.05,
             y=0.1,
             steps=[],
         )
 
-        for a in self.df["age"]:
+        # get start and end points for bound line plots
+        min_x = min(self.df["x"])
+        max_x = max(self.df["x"])
+        min_y = min(self.df["y"])
+        max_y = max(self.df["y"])
+
+        maxage = max(self.df["age"])
+        for a in np.linspace(1, maxage, min(200, maxage)):
+            a = int(a)
             adf = self.df[self.df["age"] <= a]
             node_object = go.Scatter3d(
                 x=adf["x"],
@@ -295,12 +313,34 @@ class TreeD:
                 mode="markers+text",
                 marker=marker,
                 hovertext=adf["number"],
-                hovertemplate="LP obj: %{z}<br>node number: %{hovertext}<br>%{marker.color}",
+                # hovertemplate="LP obj: %{z}<br>node number: %{hovertext}<br>%{marker.color}",
                 hoverinfo="z+text+name",
                 opacity=0.7,
-                name="LP solutions",
+                name="LP Solutions",
             )
-            frames.append(go.Frame(data=node_object, name=str(a)))
+
+            primalbound = go.Scatter3d(
+                x=[min_x, min_x, max_x, max_x, min_x],
+                y=[min_y, max_y, max_y, min_y, min_y],
+                z=[adf["primalbound"].iloc[-1]] * 5,
+                mode="lines",
+                line=go.scatter3d.Line(width=5),
+                hoverinfo="name+z",
+                name="Primal Bound",
+                opacity=0.5,
+            )
+            dualbound = go.Scatter3d(
+                x=[min_x, min_x, max_x, max_x, min_x],
+                y=[min_y, max_y, max_y, min_y, min_y],
+                z=[adf["dualbound"].iloc[-1]] * 5,
+                mode="lines",
+                line=go.scatter3d.Line(width=5),
+                hoverinfo="name+z",
+                name="Dual Bound",
+                opacity=0.5,
+            )
+
+            frames.append(go.Frame(data=[node_object, primalbound, dualbound], name=str(a)))
 
             slider_step = {
                 "args": [
@@ -333,34 +373,62 @@ class TreeD:
             active=0,
             yanchor="top",
             xanchor="left",
-            currentvalue={"prefix": "Age:", "visible": True, "xanchor": "right",},
+            currentvalue={
+                "prefix": "Age:",
+                "visible": True,
+                "xanchor": "right",
+            },
             len=0.9,
             x=0.05,
             y=0.1,
             steps=[],
         )
 
-        for a in self.df["age"]:
+        # get start and end points for bound line plots
+        xmin = min([self.pos2d[k][0] for k in self.pos2d])
+        xmax = max([self.pos2d[k][0] for k in self.pos2d])
+
+        maxage = max(self.df["age"])
+        for a in np.linspace(1, maxage, min(200, maxage)):
+            a = int(a)
             adf = self.df[self.df["age"] <= a]
             node_object = go.Scatter(
                 x=[self.pos2d[k][0] for k in adf["id"]],
-                y=[self.pos2d[k][1] for k in adf["id"]],
+                # y=[self.pos2d[k][1] for k in adf["id"]],
+                y=[self.df["objval"][k] for k in adf["id"]],
                 mode="markers",
                 marker=marker,
-                hovertext=[
-                    f"LP obj: {adf['objval'].iloc[i]:.3f}\
-                    <br>node number: {adf['number'].iloc[i]}\
-                    <br>node age: {adf['age'].iloc[i]}\
-                    <br>depth: {adf['depth'].iloc[i]}\
-                    <br>LP cond: {adf['condition'].iloc[i]:.1f}\
-                    <br>iterations: {adf['iterations'].iloc[i]}"
-                    for i in range(len(adf))
-                ],
+                # hovertext=[
+                #     f"LP obj: {adf['objval'].iloc[i]:.3f}\
+                #     <br>node number: {adf['number'].iloc[i]}\
+                #     <br>node age: {adf['age'].iloc[i]}\
+                #     <br>depth: {adf['depth'].iloc[i]}\
+                #     <br>LP cond: {adf['condition'].iloc[i]:.1f}\
+                #     <br>iterations: {adf['iterations'].iloc[i]}"
+                #     for i in range(len(adf))
+                # ],
                 hoverinfo="text+name",
                 opacity=0.7,
-                name="LP solutions",
+                name="LP Solutions",
+            )
+            primalbound = go.Scatter(
+                x=[xmin, xmax],
+                y=2 * [adf["primalbound"].iloc[-1]],
+                mode="lines",
+                opacity=0.5,
+                name="Primal Bound",
+            )
+            dualbound = go.Scatter(
+                x=[xmin, xmax],
+                y=2 * [adf["dualbound"].iloc[-1]],
+                mode="lines",
+                opacity=0.5,
+                name="Dual Bound",
+            )
+
+            frames.append(
+                go.Frame(data=[node_object, primalbound, dualbound], name=str(a))
             )
-            frames.append(go.Frame(data=node_object, name=str(a)))
 
             slider_step = {
                 "args": [
@@ -428,6 +496,24 @@ class TreeD:
                                     }
                                 ],
                             ),
+                            # dict(
+                            #     label="logarithmic obj",
+                            #     method="relayout",
+                            #     args=[
+                            #         {
+                            #             "yaxis.type": 'log'
+                            #         }
+                            #     ],
+                            # ),
+                            # dict(
+                            #     label="linear obj",
+                            #     method="relayout",
+                            #     args=[
+                            #         {
+                            #             "yaxis.type": 'linear'
+                            #         }
+                            #     ],
+                            # ),
                         ]
                     ),
                     direction="down",
@@ -438,21 +524,24 @@ class TreeD:
                     buttons=list(
                         [
                             dict(
-                                label="▶",
+                                label="|▶",
                                 method="animate",
                                 args=[
                                     None,
                                     {
-                                        "frame": {"duration": 50, "redraw": True,},
+                                        "frame": {
+                                            "duration": 50,
+                                            "redraw": True,
+                                        },
                                         "fromcurrent": True,
                                     },
                                 ],
                                 args2=[
                                     [None],
                                     {
-                                        "frame": {"duration": 0, "redraw": False},
+                                        "frame": {"duration": 1, "redraw": False},
                                         "mode": "immediate",
-                                        "transition": {"duration": 0},
+                                        "transition": {"duration": 1},
                                     },
                                 ],
                             )
@@ -488,7 +577,7 @@ class TreeD:
             mode="lines",
             line=go.scatter3d.Line(color="rgb(75,75,75)", width=2),
             hoverinfo="none",
-            name="edges",
+            name="Edges",
         )
 
         min_x = min(self.df["x"])
@@ -496,14 +585,34 @@ class TreeD:
         min_y = min(self.df["y"])
         max_y = max(self.df["y"])
 
+        primalbound = go.Scatter3d(
+            x=[min_x, min_x, max_x, max_x, min_x],
+            y=[min_y, max_y, max_y, min_y, min_y],
+            z=5 * [self.df["primalbound"].iloc[-1]],
+            mode="lines",
+            line=go.scatter3d.Line(width=5),
+            hoverinfo="name+z",
+            opacity=0.5,
+            name="Primal Bound",
+        )
+        dualbound = go.Scatter3d(
+            x=[min_x, min_x, max_x, max_x, min_x],
+            y=[min_y, max_y, max_y, min_y, min_y],
+            z=5 * [self.df["dualbound"].iloc[-1]],
+            mode="lines",
+            line=go.scatter3d.Line(width=5),
+            hoverinfo="name+z",
+            opacity=0.5,
+            name="Dual Bound",
+        )
         optval = go.Scatter3d(
             x=[min_x, min_x, max_x, max_x, min_x],
             y=[min_y, max_y, max_y, min_y, min_y],
             z=[self.optval] * 5,
             mode="lines",
-            line=go.scatter3d.Line(color="rgb(0,200,50)", width=5),
+            line=go.scatter3d.Line(width=5),
             hoverinfo="name+z",
-            name="optimal value",
+            name="Optimum",
             opacity=0.5,
         )
 
@@ -520,30 +629,37 @@ class TreeD:
             gridcolor="lightgray",
         )
         zaxis = go.layout.scene.ZAxis(
-            title="objective value", backgroundcolor="white", gridcolor="lightgray"
+            title="Objective value", backgroundcolor="white", gridcolor="lightgray"
         )
         scene = go.layout.Scene(xaxis=xaxis, yaxis=yaxis, zaxis=zaxis)
-        title = (
-            "TreeD: " + self.probname + ", " + self.scipversion if self.title else ""
-        )
+
+        if self.title:
+            title = f"TreeD: {self.probname} ({self.scipversion}, {self.status})"
+        else:
+            title = ""
+        
         filename = "TreeD_" + self.probname + ".html"
 
         layout = go.Layout(
             title=title,
             font=dict(size=self.fontsize),
+            font_family="Fira Sans",
             autosize=True,
             # width=900,
             # height=600,
             showlegend=self.showlegend,
             hovermode="closest",
             scene=scene,
+            template="none",
         )
 
         layout["updatemenus"] = self.updatemenus()
         layout["sliders"] = [sliders]
 
         self.fig = go.Figure(
-            data=[nodes, nodeprojs, edges, optval], layout=layout, frames=frames,
+            data=[nodes, primalbound, dualbound, optval, nodeprojs, edges],
+            layout=layout,
+            frames=frames,
         )
 
         self.fig.write_html(file=filename, include_plotlyjs=self.include_plotlyjs)
@@ -565,12 +681,14 @@ class TreeD:
         frames, sliders = self._create_nodes_frames_2d()
 
         Xv = [self.pos2d[k][0] for k in self.df["id"]]
-        Yv = [self.pos2d[k][1] for k in self.df["id"]]
+        # Yv = [self.pos2d[k][1] for k in self.df["id"]]
+        Yv = self.df["objval"]
         Xed = []
         Yed = []
         for edge in self.nxgraph.edges:
             Xed += [self.pos2d[edge[0]][0], self.pos2d[edge[1]][0], None]
-            Yed += [self.pos2d[edge[0]][1], self.pos2d[edge[1]][1], None]
+            # Yed += [self.pos2d[edge[0]][1], self.pos2d[edge[1]][1], None]
+            Yed += [self.df["objval"][edge[0]], self.df["objval"][edge[1]], None]
 
         colorbar = go.scatter.marker.ColorBar(title="", thickness=10, x=-0.05)
         marker = go.scatter.Marker(
@@ -587,7 +705,7 @@ class TreeD:
             mode="lines",
             line=dict(color="rgb(75,75,75)", width=1),
             hoverinfo="none",
-            name="edges",
+            name="Edges",
         )
         nodes = go.Scatter(
             x=Xv,
@@ -607,19 +725,46 @@ class TreeD:
             hoverinfo="text+name",
         )
 
-        xaxis = go.layout.XAxis(title="", visible=False)
-        yaxis = go.layout.YAxis(title="", visible=False)
-
-        title = (
-            "Tree 2D: " + self.probname + ", " + self.scipversion if self.title else ""
+        xmin = min(Xv)
+        xmax = max(Xv)
+        primalbound = go.Scatter(
+            x=[xmin, xmax],
+            y=2 * [self.df["primalbound"].iloc[len(self.df) - 1]],
+            mode="lines",
+            opacity=0.5,
+            name="Primal Bound",
         )
+        dualbound = go.Scatter(
+            x=[xmin, xmax],
+            y=2 * [self.df["dualbound"].iloc[len(self.df) - 1]],
+            mode="lines",
+            opacity=0.5,
+            name="Dual Bound",
+        )
+        optval = go.Scatter(
+            x=[xmin, xmax],
+            y=2 * [self.optval],
+            mode="lines",
+            opacity=0.5,
+            name="Optimum",
+        )
+
+        margin = 0.05*xmax
+        xaxis = go.layout.XAxis(title="", visible=False, range=[xmin-margin, xmax+margin], autorange=False)
+        yaxis = go.layout.YAxis(title="Objective value", visible=True, side="right", position=0.98)
+
+        if self.title:
+            title = f"Tree 2D: {self.probname} ({self.scipversion}, {self.status})"
+        else:
+            title = ""
         filename = "Tree_2D_" + self.probname + ".html"
 
         layout = go.Layout(
             title=title,
             font=dict(size=self.fontsize),
+            font_family="Fira Sans",
             autosize=True,
-            template="simple_white",
+            template="none",
             showlegend=self.showlegend,
             hovermode="closest",
             xaxis=xaxis,
@@ -627,10 +772,13 @@ class TreeD:
         )
 
         layout["updatemenus"] = self.updatemenus()
-
         layout["sliders"] = [sliders]
 
-        self.fig2d = go.Figure(data=[nodes, edges], layout=layout, frames=frames,)
+        self.fig2d = go.Figure(
+            data=[nodes, primalbound, dualbound, optval, edges],
+            layout=layout,
+            frames=frames,
+        )
 
         self.fig2d.write_html(file=filename, include_plotlyjs=self.include_plotlyjs)
 
@@ -645,7 +793,9 @@ class TreeD:
 
         model = Model("TreeD")
 
-        if not self.verbose:
+        if self.verbose:
+            model.redirectOutput()
+        else:
             model.hideOutput()
 
         eventhdlr = LPstatEventhdlr()
@@ -676,7 +826,8 @@ class TreeD:
         self.scipversion = "SCIP " + str(model.version())
         # self.scipversion = self.scipversion[:-1]+'.'+self.scipversion[-1]
 
-        if model.getStatus() == "optimal":
+        self.status = model.getStatus()
+        if self.status == "optimal":
             self.optval = model.getObjVal()
         else:
             self.optval = None
@@ -778,4 +929,3 @@ class TreeD:
         for k in set([*p1.keys(), *p2.keys()]):
             dist += (p1.get(k, 0) - p2.get(k, 0)) ** 2
         return math.sqrt(dist)
-
